;; Generated package description from elfeed.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "elfeed" "3.4.2.0.20260520.190044" "An Emacs Atom/RSS feed reader" '((emacs "28.1") (compat "31")) :commit "a359945e3ca9a522d50f4f704750a17438628431" :authors '(("Christopher Wellons" . "wellons@nullprogram.com")) :maintainer '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com") ("Ihor Radchenko" . "yantar92@posteo.net") ("Daniel Mendler" . "mail@daniel-mendler.de")) :keywords '("network" "comm" "hypermedia") :url "https://github.com/emacs-elfeed/elfeed")
